EnableExplicit

; #################################################### Documentation #############################################
; 
; D3HT - D3 HashTable
;
; Version History:
; - 0.000 05.03.2011
; 
; - 1.000 05.03.2011
;     - Implemented everything
; 
; 
; 
; Documentation:
; - Table-Element:
;     - Metadata  (1 Byte)
;     - Key       (Depends on the Key_Size)
;     - Value     (Depends on the Value_Size)
; 
; #################################################### Constants #################################################

#D3HT_Version = 1000

#D3HT_Result_Fail = 0
#D3HT_Result_Success = 1

#D3HT_Buffer_Elements_Default = 512
#D3HT_Sidesearch_Deep_Default = 2

; #################################################### Structures / Variables ####################################

Structure D3HT_Table_Buffer
  Start_Hash.i
  
  *Memory
  
  Elements.i
EndStructure

Structure D3HT_Table
  Element_Key_Size.i    ; Size of the Key of each Element
  Element_Value_Size.i  ; Size of the Data of each Element
  
  Hash_Mask.i           ; Mask for the hash, depends on Buffer_Elements
  
  Elements.i            ; Amount of elements
  Buffer_Elements.i     ; Elements per Buffer
  Sidesearch_Deep.i     ; Amount of iterations to search "sideways" (Slower, but more memory friendly)
  List Buffer.D3HT_Table_Buffer()
EndStructure

; #################################################### Declares ##################################################

; #################################################### Macros ####################################################

; #################################################### Procedures ################################################

Procedure D3HT_Key_2_Hash(*Key.Ascii, Key_Size, Start_Hash, Output_Mask)
  Protected i
  
  For i = 1 To Key_Size
    Start_Hash = *Key\a + (Start_Hash << 6) + (Start_Hash >> 16) - Start_Hash
    *Key + 1
  Next
  
  ;Start_Hash = (Start_Hash << 3) + (Start_Hash >> 5) - Start_Hash
  ;Start_Hash = (Start_Hash << 3) + (Start_Hash >> 5) - Start_Hash
  ;Start_Hash = (Start_Hash << 3) + (Start_Hash >> 5) - Start_Hash
  ;Start_Hash = (Start_Hash << 3) + (Start_Hash >> 5) - Start_Hash
  
  ProcedureReturn Start_Hash & Output_Mask
EndProcedure

Procedure D3HT_Create(Key_Size, Value_Size, Buffer_Elements=#D3HT_Buffer_Elements_Default, Sidesearch_Deep=#D3HT_Sidesearch_Deep_Default)
  Protected *D3HT_Table.D3HT_Table
  
  Select Buffer_Elements
    Case 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536
    Default : ProcedureReturn #D3HT_Result_Fail
  EndSelect
  
  *D3HT_Table = AllocateMemory(SizeOf(D3HT_Table))
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  InitializeStructure(*D3HT_Table, D3HT_Table)
  
  *D3HT_Table\Hash_Mask = Buffer_Elements-1
  
  *D3HT_Table\Buffer_Elements = Buffer_Elements
  *D3HT_Table\Sidesearch_Deep = Sidesearch_Deep
  *D3HT_Table\Element_Key_Size = Key_Size
  *D3HT_Table\Element_Value_Size = Value_Size
  
  ProcedureReturn *D3HT_Table
EndProcedure

Procedure D3HT_Destroy(*D3HT_Table.D3HT_Table)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  While LastElement(*D3HT_Table\Buffer())
    If *D3HT_Table\Buffer()\Memory
      FreeMemory(*D3HT_Table\Buffer()\Memory)
    EndIf
    DeleteElement(*D3HT_Table\Buffer())
  Wend
  
  ClearStructure(*D3HT_Table, D3HT_Table)
  FreeMemory(*D3HT_Table)
  
  ProcedureReturn #D3HT_Result_Success
EndProcedure

Procedure D3HT_Clear(*D3HT_Table.D3HT_Table)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  While LastElement(*D3HT_Table\Buffer())
    If *D3HT_Table\Buffer()\Memory
      FreeMemory(*D3HT_Table\Buffer()\Memory)
    EndIf
    DeleteElement(*D3HT_Table\Buffer())
  Wend
  
  *D3HT_Table\Elements = 0
  
  ProcedureReturn #D3HT_Result_Success
EndProcedure

Procedure D3HT_Get_Elements(*D3HT_Table.D3HT_Table)
  If Not *D3HT_Table
    ProcedureReturn -1 ; Invalid amount of elements
  EndIf
  
  ProcedureReturn *D3HT_Table\Elements
EndProcedure

Procedure D3HT_Get_Memoryusage(*D3HT_Table.D3HT_Table)
  Protected Memoryusage = SizeOf(D3HT_Table)
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ForEach *D3HT_Table\Buffer()
    Memoryusage + SizeOf(D3HT_Table_Buffer)
    Memoryusage + MemorySize(*D3HT_Table\Buffer()\Memory)
  Next
  
  ProcedureReturn Memoryusage
EndProcedure

Procedure D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, *Value, Check_Collision=1)
  Protected Element_Pos
  Protected *Pointer.Ascii
  Protected Sidesearch_Iteration
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *Key
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *Value
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Check_Collision
    ; #### Look if the element already exsists
    ForEach *D3HT_Table\Buffer()
      Element_Pos = D3HT_Key_2_Hash(*Key, *D3HT_Table\Element_Key_Size, *D3HT_Table\Buffer()\Start_Hash, *D3HT_Table\Hash_Mask)
      
      Sidesearch_Iteration = *D3HT_Table\Sidesearch_Deep + 1
      Repeat
        *Pointer = *D3HT_Table\Buffer()\Memory + Element_Pos * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size)
        If *Pointer\a & $01 ; If element is used
          If CompareMemory(*Key, *Pointer+1, *D3HT_Table\Element_Key_Size)
            ; #### Found element, overwrite its value now
            MoveMemory(*Value, *Pointer+1+*D3HT_Table\Element_Key_Size, *D3HT_Table\Element_Value_Size)
            ProcedureReturn #D3HT_Result_Success
          EndIf
        EndIf
        
        Element_Pos + 1
        Sidesearch_Iteration - 1
      Until Sidesearch_Iteration = 0 Or Element_Pos = *D3HT_Table\Buffer_Elements
    Next
  EndIf
  
  ; #### Haven't found the element, so create a new one
  ForEach *D3HT_Table\Buffer()
    Element_Pos = D3HT_Key_2_Hash(*Key, *D3HT_Table\Element_Key_Size, *D3HT_Table\Buffer()\Start_Hash, *D3HT_Table\Hash_Mask)
    
    Sidesearch_Iteration = *D3HT_Table\Sidesearch_Deep + 1
    Repeat
      *Pointer = *D3HT_Table\Buffer()\Memory + Element_Pos * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size)
      If Not *Pointer\a & $01 ; If element is free
        ; #### Found free element, write its key and value now
        *Pointer\a | $01
        *D3HT_Table\Buffer()\Elements + 1
        *D3HT_Table\Elements + 1
        MoveMemory(*Key, *Pointer+1, *D3HT_Table\Element_Key_Size)
        MoveMemory(*Value, *Pointer+1+*D3HT_Table\Element_Key_Size, *D3HT_Table\Element_Value_Size)
        ProcedureReturn #D3HT_Result_Success
      EndIf
      
      Element_Pos + 1
      Sidesearch_Iteration - 1
    Until Sidesearch_Iteration = 0 Or Element_Pos = *D3HT_Table\Buffer_Elements
  Next
  
  ; #### Haven't even found a free element in existing buffers, so create a new buffer...
  LastElement(*D3HT_Table\Buffer())
  If Not AddElement(*D3HT_Table\Buffer())
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  *D3HT_Table\Buffer()\Memory = AllocateMemory(*D3HT_Table\Buffer_Elements * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size))
  If Not *D3HT_Table\Buffer()\Memory
    DeleteElement(*D3HT_Table\Buffer())
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  *D3HT_Table\Buffer()\Start_Hash = Random(65535)
  
  Element_Pos = D3HT_Key_2_Hash(*Key, *D3HT_Table\Element_Key_Size, *D3HT_Table\Buffer()\Start_Hash, *D3HT_Table\Hash_Mask)
  
  *Pointer = *D3HT_Table\Buffer()\Memory + Element_Pos * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size)

  ; #### write its key and value now
  *Pointer\a | $01
  *D3HT_Table\Buffer()\Elements + 1
  *D3HT_Table\Elements + 1
  MoveMemory(*Key, *Pointer+1, *D3HT_Table\Element_Key_Size)
  MoveMemory(*Value, *Pointer+1+*D3HT_Table\Element_Key_Size, *D3HT_Table\Element_Value_Size)
  ProcedureReturn #D3HT_Result_Success
EndProcedure

Procedure D3HT_Element_Set_Byte(*D3HT_Table.D3HT_Table, *Key, Value.b, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 1
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Ascii(*D3HT_Table.D3HT_Table, *Key, Value.a, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 1
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Word(*D3HT_Table.D3HT_Table, *Key, Value.w, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 2
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Unicode(*D3HT_Table.D3HT_Table, *Key, Value.u, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 2
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Long(*D3HT_Table.D3HT_Table, *Key, Value.l, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 4
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Quad(*D3HT_Table.D3HT_Table, *Key, Value.q, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 8
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Integer(*D3HT_Table.D3HT_Table, *Key, Value.i, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = SizeOf(Integer)
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Float(*D3HT_Table.D3HT_Table, *Key, Value.f, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 4
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Set_Double(*D3HT_Table.D3HT_Table, *Key, Value.d, Check_Collision=1)
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 8
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ProcedureReturn D3HT_Element_Set(*D3HT_Table.D3HT_Table, *Key, @Value, Check_Collision)
EndProcedure

Procedure D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, *Value)
  Protected Element_Pos
  Protected *Pointer.Ascii
  Protected Sidesearch_Iteration
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *Key
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *Value
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ; #### Look if the element already exsists
  ForEach *D3HT_Table\Buffer()
    Element_Pos = D3HT_Key_2_Hash(*Key, *D3HT_Table\Element_Key_Size, *D3HT_Table\Buffer()\Start_Hash, *D3HT_Table\Hash_Mask)
    
    Sidesearch_Iteration = *D3HT_Table\Sidesearch_Deep + 1
    Repeat
      *Pointer = *D3HT_Table\Buffer()\Memory + Element_Pos * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size)
      If *Pointer\a & $01 ; If element is used
        If CompareMemory(*Key, *Pointer+1, *D3HT_Table\Element_Key_Size)
          ; #### Found element, return its value now
          MoveMemory(*Pointer+1+*D3HT_Table\Element_Key_Size, *Value, *D3HT_Table\Element_Value_Size)
          ProcedureReturn #D3HT_Result_Success
        EndIf
      EndIf
      
      Element_Pos + 1
      Sidesearch_Iteration - 1
    Until Sidesearch_Iteration = 0 Or Element_Pos = *D3HT_Table\Buffer_Elements
  Next
  
  ProcedureReturn #D3HT_Result_Fail
EndProcedure

Procedure.b D3HT_Element_Get_Byte(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.b
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 1
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.a D3HT_Element_Get_Ascii(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.a
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 1
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.w D3HT_Element_Get_Word(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.w
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 2
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.u D3HT_Element_Get_Unicode(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.u
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 2
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.l D3HT_Element_Get_Long(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.l
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 4
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.q D3HT_Element_Get_Quad(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.q
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 8
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.i D3HT_Element_Get_Integer(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.i
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = SizeOf(Integer)
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value
EndProcedure

Procedure.f D3HT_Element_Get_Float(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.f
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 4
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value 
EndProcedure

Procedure.d D3HT_Element_Get_Double(*D3HT_Table.D3HT_Table, *Key)
  Protected Value.d
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *D3HT_Table\Element_Value_Size = 8
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  D3HT_Element_Get(*D3HT_Table.D3HT_Table, *Key, @Value)
  
  ProcedureReturn Value 
EndProcedure

Procedure D3HT_Element_Free(*D3HT_Table.D3HT_Table, *Key)
  Protected Element_Pos
  Protected *Pointer.Ascii
  Protected Sidesearch_Iteration
  
  If Not *D3HT_Table
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  If Not *Key
    ProcedureReturn #D3HT_Result_Fail
  EndIf
  
  ; #### Look if the element already exsists
  ForEach *D3HT_Table\Buffer()
    Element_Pos = D3HT_Key_2_Hash(*Key, *D3HT_Table\Element_Key_Size, *D3HT_Table\Buffer()\Start_Hash, *D3HT_Table\Hash_Mask)
    
    Sidesearch_Iteration = *D3HT_Table\Sidesearch_Deep + 1
    Repeat
      *Pointer = *D3HT_Table\Buffer()\Memory + Element_Pos * (1 + *D3HT_Table\Element_Key_Size + *D3HT_Table\Element_Value_Size)
      If *Pointer\a & $01 ; If element is used
        If CompareMemory(*Key, *Pointer+1, *D3HT_Table\Element_Key_Size)
          ; #### Found element, delete it
          *Pointer\a & ~$01
          *D3HT_Table\Buffer()\Elements - 1
          *D3HT_Table\Elements - 1
          If *D3HT_Table\Buffer()\Elements = 0
            ; #### No more elements in the buffer, delete it
            FreeMemory(*D3HT_Table\Buffer()\Memory)
            DeleteElement(*D3HT_Table\Buffer())
          EndIf
          ProcedureReturn #D3HT_Result_Success
        EndIf
      EndIf
      
      Element_Pos + 1
      Sidesearch_Iteration - 1
    Until Sidesearch_Iteration = 0 Or Element_Pos = *D3HT_Table\Buffer_Elements
  Next
  
  ProcedureReturn #D3HT_Result_Fail
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 133
; FirstLine = 116
; Folding = -----
; EnableXP
; DisableDebugger